import time

from azure.cognitiveservices.vision.computervision import ComputerVisionClient
from azure.cognitiveservices.vision.computervision.models import OperationStatusCodes
from loguru import logger
from msrest.authentication import CognitiveServicesCredentials

from neco.ocr.base_ocr import BaseOCR


class AzureOCR(BaseOCR):
    def __init__(self, azure_ocr_endpoint: str, azure_ocr_key: str):
        self.azure_ocr_endpoint = azure_ocr_endpoint
        self.azure_ocr_key = azure_ocr_key

    def predict(self, file) -> str:
        logger.info(f"使用Azure OCR识别图片:[{file}]")
        with open(file, "rb") as image:
            try:
                computervision_client = ComputerVisionClient(self.azure_ocr_endpoint,
                                                             CognitiveServicesCredentials(self.azure_ocr_key))
                read_response = computervision_client.read_in_stream(
                    image, raw=True)
                read_operation_location = read_response.headers["Operation-Location"]
                operation_id = read_operation_location.split("/")[-1]
                while True:
                    read_result = computervision_client.get_read_result(
                        operation_id)
                    if read_result.status not in ['notStarted', 'running']:
                        break
                    time.sleep(1)

                content = ''
                if read_result.status == OperationStatusCodes.succeeded:
                    for text_result in read_result.analyze_result.read_results:
                        for line in text_result.lines:
                            content += line.text + ' '
            except Exception as ex:
                logger.error(f"Azure OCR识别图片失败: {ex}")
                content = ''
        return content



from neco.mlops.anomaly_detection.base_anomaly_detection import BaseAnomalyDetection
from sklearn.ensemble import RandomForestClassifier


class RandomForestAnomalyDetector(BaseAnomalyDetection):
    def build_model(self, train_params):
        model = RandomForestClassifier(**train_params)
        return model



import os
import torch
import numpy as np
from tqdm import tqdm
from typing import List, Dict, Tuple, Type, Union, Optional, Any
from copy import deepcopy
from transformers import AutoTokenizer, AutoModelForSequenceClassification
from loguru import logger


# 常量定义
DEFAULT_MODEL_NAME = 'maidalun1020/bce-reranker-base_v1'
DEFAULT_BATCH_SIZE = 256
DEFAULT_MAX_LENGTH = 512
DEFAULT_OVERLAP_TOKENS = 80
MAX_PASSAGE_LENGTH = 128000  # 最大文档长度限制
MIN_QUERY_LENGTH = 1
MIN_PASSAGE_LENGTH = 1
MAX_QUERY_TOKENS = 400  # 查询最大token数


class BCEReRankConfig:
    """BCE重排序模型配置"""

    def __init__(
        self,
        model_name_or_path: str = DEFAULT_MODEL_NAME,
        use_fp16: bool = False,
        device: Optional[str] = None,
        cache_dir: Optional[str] = None,
        max_length: int = DEFAULT_MAX_LENGTH,
        overlap_tokens: int = DEFAULT_OVERLAP_TOKENS,
        **kwargs
    ) -> None:
        self.model_name_or_path = model_name_or_path
        self.use_fp16 = use_fp16
        self.device = device
        self.cache_dir = cache_dir or os.getenv(
            'BCE_MODEL_CACHE_DIR', './models')
        self.max_length = max_length
        self.overlap_tokens = overlap_tokens
        self.kwargs = kwargs


class BCEReRank:
    """BCE重排序模型实现

    用于对文档进行重排序，提升检索结果的相关性。
    支持GPU加速和批量处理。
    """

    def __init__(
        self,
        model_name_or_path: str = DEFAULT_MODEL_NAME,
        use_fp16: bool = False,
        device: Optional[str] = None,
        **kwargs
    ) -> None:
        """初始化BCE重排序模型

        Args:
            model_name_or_path: 模型名称或路径
            use_fp16: 是否使用半精度浮点数
            device: 指定设备，如'cpu', 'cuda', 'cuda:0'等
            **kwargs: 其他配置参数

        Raises:
            ValueError: 设备配置错误
            RuntimeError: 模型加载失败
        """
        logger.info(f"正在初始化BCE重排序模型: {model_name_or_path}")

        self.config = BCEReRankConfig(
            model_name_or_path=model_name_or_path,
            use_fp16=use_fp16,
            device=device,
            **kwargs
        )

        try:
            self._load_model()
            self._setup_device()
            self._configure_model()

            logger.info(
                f"BCE重排序模型初始化完成 - "
                f"设备: {self.device}, GPU数量: {self.num_gpus}, "
                f"FP16: {use_fp16}, 模型: {model_name_or_path}"
            )
        except Exception as e:
            logger.error(f"BCE重排序模型初始化失败: {e}", exc_info=True)
            raise RuntimeError(f"模型初始化失败: {e}") from e

    def _load_model(self) -> None:
        """加载tokenizer和模型"""
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.config.model_name_or_path,
                cache_dir=self.config.cache_dir,
                **self.config.kwargs
            )
            self.model = AutoModelForSequenceClassification.from_pretrained(
                self.config.model_name_or_path,
                cache_dir=self.config.cache_dir,
                **self.config.kwargs
            )
        except Exception as e:
            logger.error(f"模型加载失败: {e}")
            raise

    def _setup_device(self) -> None:
        """设置计算设备"""
        num_gpus = torch.cuda.device_count()

        if self.config.device is None:
            self.device = "cuda" if num_gpus > 0 else "cpu"
        else:
            device_str = self.config.device
            if device_str.isdigit():
                self.device = f'cuda:{int(device_str)}'
            else:
                self.device = device_str

        # 验证设备有效性
        if self.device not in ["cpu", "cuda"] and not self.device.startswith('cuda:'):
            raise ValueError(
                f"无效设备配置: {self.device}. 支持的设备: 'cpu', 'cuda', 'cuda:0'等")

        # 设置GPU数量
        if self.device == "cpu":
            self.num_gpus = 0
        elif self.device.startswith('cuda:') and num_gpus > 0:
            self.num_gpus = 1
        elif self.device == "cuda":
            self.num_gpus = num_gpus
        else:
            logger.warning(f"设备 {self.device} 不可用，回退到CPU")
            self.device = "cpu"
            self.num_gpus = 0

    def _configure_model(self) -> None:
        """配置模型"""
        if self.config.use_fp16:
            self.model.half()

        self.model.eval()
        self.model = self.model.to(self.device)

        if self.num_gpus > 1:
            self.model = torch.nn.DataParallel(self.model)

    def compute_score(
        self,
        sentence_pairs: Union[List[Tuple[str, str]], Tuple[str, str]],
        batch_size: int = DEFAULT_BATCH_SIZE,
        max_length: int = DEFAULT_MAX_LENGTH,
        enable_tqdm: bool = True,
        **kwargs
    ) -> Union[float, List[float]]:
        """计算句子对的相似度分数

        Args:
            sentence_pairs: 句子对列表或单个句子对
            batch_size: 批处理大小
            max_length: 最大序列长度
            enable_tqdm: 是否显示进度条
            **kwargs: 其他参数

        Returns:
            相似度分数或分数列表

        Raises:
            ValueError: 输入格式错误
        """
        # 输入验证
        if not sentence_pairs:
            raise ValueError("sentence_pairs不能为空")

        if self.num_gpus > 1:
            batch_size = batch_size * self.num_gpus

        # 标准化输入格式
        if not isinstance(sentence_pairs, list):
            sentence_pairs = [sentence_pairs]

        if len(sentence_pairs) > 0 and isinstance(sentence_pairs[0], str):
            sentence_pairs = [sentence_pairs]

        logger.debug(
            f"开始计算相似度分数 - 句子对数量: {len(sentence_pairs)}, 批大小: {batch_size}")

        try:
            with torch.no_grad():
                scores_collection = []
                total_batches = (len(sentence_pairs) +
                                 batch_size - 1) // batch_size

                for sentence_id in tqdm(
                    range(0, len(sentence_pairs), batch_size),
                    desc='计算相似度分数',
                    disable=not enable_tqdm,
                    total=total_batches
                ):
                    sentence_pairs_batch = sentence_pairs[sentence_id:sentence_id + batch_size]

                    inputs = self.tokenizer(
                        sentence_pairs_batch,
                        padding=True,
                        truncation=True,
                        max_length=max_length,
                        return_tensors="pt"
                    )

                    inputs_on_device = {k: v.to(self.device)
                                        for k, v in inputs.items()}
                    scores = self.model(
                        **inputs_on_device, return_dict=True).logits.view(-1).float()
                    scores = torch.sigmoid(scores)
                    scores_collection.extend(scores.cpu().numpy().tolist())

            logger.debug(f"相似度分数计算完成 - 共处理 {len(scores_collection)} 个句子对")

            if len(scores_collection) == 1:
                return scores_collection[0]
            return scores_collection

        except Exception as e:
            logger.error(f"计算相似度分数失败: {e}", exc_info=True)
            raise

    def rerank(
        self,
        query: str,
        passages: List[str],
        batch_size: int = DEFAULT_BATCH_SIZE,
        **kwargs
    ) -> Dict[str, Any]:
        """对文档进行重排序

        Args:
            query: 查询字符串
            passages: 待重排序的文档列表
            batch_size: 批处理大小
            **kwargs: 其他参数

        Returns:
            包含重排序结果的字典，包含以下键：
            - rerank_passages: 重排序后的文档列表
            - rerank_scores: 对应的相关性分数列表
            - rerank_ids: 重排序后的原始索引列表

        Raises:
            ValueError: 输入参数无效
        """
        logger.debug(
            f"开始重排序 - 查询长度: {len(query) if query else 0}, 文档数量: {len(passages)}")

        # 输入验证和预处理
        if not self._validate_rerank_inputs(query, passages):
            return {'rerank_passages': [], 'rerank_scores': [], 'rerank_ids': []}

        # 清理和截断文档
        cleaned_passages = self._clean_passages(passages)
        if not cleaned_passages:
            logger.warning("清理后无有效文档")
            return {'rerank_passages': [], 'rerank_scores': [], 'rerank_ids': []}

        try:
            # 预处理tokenization
            sentence_pairs, sentence_pairs_pids = self._preprocess_for_tokenization(
                query, cleaned_passages
            )

            # 批量推理
            scores = self._batch_inference(sentence_pairs, batch_size)

            # 合并分数和排序
            result = self._merge_and_sort_results(
                cleaned_passages, sentence_pairs_pids, scores)

            logger.debug(
                f"重排序完成 - 输入文档: {len(passages)}, "
                f"输出文档: {len(result['rerank_passages'])}"
            )

            return result

        except Exception as e:
            logger.error(f"重排序处理失败: {e}", exc_info=True)
            return {'rerank_passages': [], 'rerank_scores': [], 'rerank_ids': []}

    def _validate_rerank_inputs(self, query: str, passages: List[str]) -> bool:
        """验证重排序输入参数"""
        if not query or len(query.strip()) < MIN_QUERY_LENGTH:
            logger.warning("查询为空或过短")
            return False

        if not passages:
            logger.warning("文档列表为空")
            return False

        return True

    def _clean_passages(self, passages: List[str]) -> List[str]:
        """清理和验证文档列表"""
        cleaned = []
        for passage in passages:
            if isinstance(passage, str) and MIN_PASSAGE_LENGTH <= len(passage.strip()):
                # 截断过长的文档
                truncated = passage[:MAX_PASSAGE_LENGTH]
                cleaned.append(truncated)

        return cleaned

    def _preprocess_for_tokenization(
        self,
        query: str,
        passages: List[str]
    ) -> Tuple[List[Dict[str, Any]], List[int]]:
        """预处理tokenization"""
        return self.reranker_tokenize_preproc(
            query,
            passages,
            tokenizer=self.tokenizer,
            max_length=self.config.max_length,
            overlap_tokens=self.config.overlap_tokens,
        )

    def _batch_inference(self, sentence_pairs: List[Dict[str, Any]], batch_size: int) -> List[float]:
        """批量推理"""
        if self.num_gpus > 1:
            batch_size = batch_size * self.num_gpus

        tot_scores = []
        total_batches = (len(sentence_pairs) + batch_size - 1) // batch_size

        with torch.no_grad():
            for k in tqdm(
                range(0, len(sentence_pairs), batch_size),
                desc='重排序推理',
                total=total_batches
            ):
                batch = self.tokenizer.pad(
                    sentence_pairs[k:k + batch_size],
                    padding=True,
                    max_length=None,
                    pad_to_multiple_of=None,
                    return_tensors="pt"
                )
                batch_on_device = {k: v.to(self.device)
                                   for k, v in batch.items()}
                scores = self.model(**batch_on_device,
                                    return_dict=True).logits.view(-1).float()
                scores = torch.sigmoid(scores)
                tot_scores.extend(scores.cpu().numpy().tolist())

        return tot_scores

    def _merge_and_sort_results(
        self,
        passages: List[str],
        sentence_pairs_pids: List[int],
        scores: List[float]
    ) -> Dict[str, Any]:
        """合并分数并排序结果"""
        # 初始化每个文档的最大分数
        merge_scores = [0.0 for _ in range(len(passages))]

        # 对于每个文档片段，取最高分数
        for pid, score in zip(sentence_pairs_pids, scores):
            if 0 <= pid < len(merge_scores):
                merge_scores[pid] = max(merge_scores[pid], score)

        # 按分数降序排序
        merge_scores_argsort = np.argsort(merge_scores)[::-1]

        sorted_passages = []
        sorted_scores = []

        for mid in merge_scores_argsort:
            sorted_scores.append(merge_scores[mid])
            sorted_passages.append(passages[mid])

        return {
            'rerank_passages': sorted_passages,
            'rerank_scores': sorted_scores,
            'rerank_ids': merge_scores_argsort.tolist()
        }

    def reranker_tokenize_preproc(
        self,
        query: str,
        passages: List[str],
        tokenizer=None,
        max_length: int = DEFAULT_MAX_LENGTH,
        overlap_tokens: int = DEFAULT_OVERLAP_TOKENS,
    ) -> Tuple[List[Dict[str, Any]], List[int]]:
        """预处理tokenization

        将查询和文档进行tokenization预处理，支持长文档的滑动窗口切分。

        Args:
            query: 查询字符串
            passages: 文档列表
            tokenizer: tokenizer实例
            max_length: 最大序列长度
            overlap_tokens: 滑动窗口重叠token数

        Returns:
            Tuple包含：
            - 处理后的输入对列表
            - 对应的文档ID列表

        Raises:
            AssertionError: tokenizer为None
            ValueError: 查询过长
        """
        if tokenizer is None:
            raise ValueError("tokenizer不能为None")

        logger.debug(
            f"开始tokenization预处理 - 查询长度: {len(query)}, 文档数: {len(passages)}")

        try:
            sep_id = tokenizer.sep_token_id

            # 对查询进行tokenization
            query_inputs = tokenizer.encode_plus(
                query, truncation=False, padding=False)
            query_token_length = len(query_inputs['input_ids'])

            # 计算文档可用的最大token长度（减去查询tokens和分隔符）
            max_passage_tokens = max_length - query_token_length - 2

            if max_passage_tokens <= 100:
                raise ValueError(
                    f"查询过长({query_token_length} tokens)，"
                    f"请确保查询长度不超过{MAX_QUERY_TOKENS} tokens"
                )

            # 计算实际的重叠token数（不超过文档长度的1/4）
            effective_overlap = min(overlap_tokens, max_passage_tokens // 4)

            res_merge_inputs = []
            res_merge_inputs_pids = []

            for pid, passage in enumerate(passages):
                passage_inputs = tokenizer.encode_plus(
                    passage,
                    truncation=False,
                    padding=False,
                    add_special_tokens=False
                )
                passage_token_length = len(passage_inputs['input_ids'])

                if passage_token_length <= max_passage_tokens:
                    # 文档较短，直接处理
                    merged_inputs = self._merge_query_passage(
                        query_inputs, passage_inputs, sep_id)
                    res_merge_inputs.append(merged_inputs)
                    res_merge_inputs_pids.append(pid)
                else:
                    # 文档较长，使用滑动窗口
                    chunks = self._split_passage_with_sliding_window(
                        passage_inputs, max_passage_tokens, effective_overlap
                    )

                    for chunk in chunks:
                        merged_inputs = self._merge_query_passage(
                            query_inputs, chunk, sep_id)
                        res_merge_inputs.append(merged_inputs)
                        res_merge_inputs_pids.append(pid)

            logger.debug(
                f"tokenization预处理完成 - 生成输入对数: {len(res_merge_inputs)}, "
                f"文档ID数: {len(res_merge_inputs_pids)}"
            )

            return res_merge_inputs, res_merge_inputs_pids

        except Exception as e:
            logger.error(f"tokenization预处理失败: {e}", exc_info=True)
            raise

    def _merge_query_passage(
        self,
        query_inputs: Dict[str, Any],
        passage_inputs: Dict[str, Any],
        sep_id: int
    ) -> Dict[str, Any]:
        """合并查询和文档的token输入"""
        merged = deepcopy(query_inputs)

        # 构建: [CLS] query [SEP] passage [SEP]
        merged['input_ids'].append(sep_id)
        merged['input_ids'].extend(passage_inputs['input_ids'])
        merged['input_ids'].append(sep_id)

        # 扩展attention_mask
        merged['attention_mask'].append(1)  # 为第一个SEP
        merged['attention_mask'].extend(passage_inputs['attention_mask'])
        merged['attention_mask'].append(1)  # 为第二个SEP

        # 如果存在token_type_ids，为文档部分设置为1
        if 'token_type_ids' in merged:
            token_type_ids = [
                1] * (len(passage_inputs['input_ids']) + 2)  # +2 for SEPs
            merged['token_type_ids'].extend(token_type_ids)

        return merged

    def _split_passage_with_sliding_window(
        self,
        passage_inputs: Dict[str, Any],
        max_tokens: int,
        overlap_tokens: int
    ) -> List[Dict[str, Any]]:
        """使用滑动窗口分割长文档"""
        chunks = []
        passage_length = len(passage_inputs['input_ids'])
        start_idx = 0

        while start_idx < passage_length:
            end_idx = min(start_idx + max_tokens, passage_length)

            # 创建当前片段
            chunk = {k: v[start_idx:end_idx]
                     for k, v in passage_inputs.items()}
            chunks.append(chunk)

            # 计算下一个窗口的起始位置
            if end_idx >= passage_length:
                break
            start_idx = end_idx - overlap_tokens

        return chunks



from typing import Any, Dict, List, Optional, Tuple
import mlflow
import mlflow.sklearn
from loguru import logger


class MLFlowUtils:
    """MLflow 工具类，提供可复用的 MLflow 操作"""
    
    @staticmethod
    def setup_experiment(tracking_uri: str, experiment_name: str):
        """设置 MLflow 实验
        
        Args:
            tracking_uri: MLflow tracking 服务地址，如果为 None 则使用本地文件系统
            experiment_name: 实验名称
        """
        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)
        else:
            mlflow.set_tracking_uri("file:./mlruns")
        mlflow.set_experiment(experiment_name)
    
    @staticmethod
    def load_model_with_metadata(
        model_name: str,
        model_version: str = "latest",
        metadata_attrs: Optional[List[str]] = None
    ) -> Tuple[Any, Dict[str, Any]]:
        """从 MLflow 模型注册中心加载模型及元数据
        
        Args:
            model_name: 模型名称
            model_version: 模型版本，默认为 "latest"
            metadata_attrs: 需要提取的元数据属性列表
            
        Returns:
            (model, metadata): 模型对象和元数据字典
            
        Raises:
            ValueError: 如果模型缺少必要的元数据属性
        """
        model_uri = f"models:/{model_name}/{model_version}"
        logger.info(f"正在加载模型: {model_uri}")
        
        model = mlflow.sklearn.load_model(model_uri)
        
        # 提取元数据
        metadata = {}
        if metadata_attrs:
            for attr in metadata_attrs:
                # 使用哨兵值来区分"属性不存在"和"属性值为None"
                sentinel = object()
                value = getattr(model, attr, sentinel)
                
                if value is sentinel:
                    logger.warning(f"模型缺少元数据属性: {attr} (属性不存在)")
                    metadata[attr] = None
                else:
                    metadata[attr] = value
                    if value is None:
                        logger.debug(f"元数据属性 {attr} 的值为 None (这可能是预期行为)")
        
        logger.info(f"模型加载完成，包含 {len(metadata)} 个元数据属性")
        return model, metadata
    
    @staticmethod
    def log_params_batch(params: Dict[str, Any]):
        """批量记录参数到 MLflow
        
        Args:
            params: 参数字典
        """
        if params:
            mlflow.log_params(params)
            logger.debug(f"已记录 {len(params)} 个参数")
    
    @staticmethod
    def log_metrics_batch(metrics: Dict[str, float], prefix: str = "", step: Optional[int] = None):
        """批量记录指标到 MLflow
        
        Args:
            metrics: 指标字典
            prefix: 指标名称前缀，如 "train_", "val_", "test_"
            step: 记录步骤（用于时间序列指标）
        """
        if metrics:
            prefixed_metrics = {f"{prefix}{k}": v for k, v in metrics.items()}
            if step is not None:
                for key, value in prefixed_metrics.items():
                    mlflow.log_metric(key, value, step=step)
            else:
                mlflow.log_metrics(prefixed_metrics)
            logger.debug(f"已记录 {len(metrics)} 个指标 (prefix={prefix})")
    
    @staticmethod
    def log_training_history(
        train_scores: List[float],
        val_scores: List[float],
        metric_name: str,
        additional_metrics: Optional[Dict[str, List[float]]] = None
    ):
        """记录训练历史到 MLflow
        
        Args:
            train_scores: 训练集分数列表
            val_scores: 验证集分数列表
            metric_name: 指标名称，如 "f1", "accuracy"
            additional_metrics: 其他指标字典，如 {"overfitting": [...]}
        """
        for i, (train_score, val_score) in enumerate(zip(train_scores, val_scores), 1):
            mlflow.log_metric(f"train_{metric_name}", train_score, step=i)
            mlflow.log_metric(f"val_{metric_name}", val_score, step=i)
            
            # 记录过拟合程度
            overfitting = train_score - val_score
            mlflow.log_metric("overfitting", overfitting, step=i)
        
        # 记录其他指标
        if additional_metrics:
            for metric_key, metric_values in additional_metrics.items():
                for i, value in enumerate(metric_values, 1):
                    mlflow.log_metric(metric_key, value, step=i)
        
        logger.info(f"已记录 {len(train_scores)} 步的训练历史")
    
    @staticmethod
    def log_model_sklearn(
        model: Any,
        registered_model_name: str,
        artifact_path: str = "model"
    ) -> str:
        """记录 sklearn 模型到 MLflow
        
        Args:
            model: sklearn 模型对象
            registered_model_name: 注册模型的名称
            artifact_path: 模型存储路径
            
        Returns:
            模型 URI
        """
        model_info = mlflow.sklearn.log_model(
            sk_model=model,
            artifact_path=artifact_path,
            registered_model_name=registered_model_name
        )
        logger.info(f"模型已注册: {registered_model_name}")
        return model_info.model_uri
    
    @staticmethod
    def get_latest_model_version(model_name: str) -> Optional[str]:
        """获取模型的最新版本号
        
        Args:
            model_name: 模型名称
            
        Returns:
            最新版本号，如果模型不存在返回 None
        """
        try:
            client = mlflow.tracking.MlflowClient()
            latest_versions = client.get_latest_versions(model_name, stages=["None", "Production", "Staging"])
            if latest_versions:
                versions = [int(v.version) for v in latest_versions]
                return str(max(versions))
            return None
        except Exception as e:
            logger.warning(f"获取模型最新版本失败: {e}")
            return None
    
    @staticmethod
    def log_training_results(
        params: Dict[str, Any],
        train_scores: List[float],
        val_scores: List[float],
        metric_name: str,
        val_metrics: Optional[Dict[str, float]] = None,
        test_metrics: Optional[Dict[str, float]] = None,
        model: Optional[Any] = None,
        model_name: Optional[str] = None
    ):
        """一站式记录训练结果到 MLflow
        
        这是一个便捷方法，集成了参数、训练历史、评估指标和模型的记录。
        
        Args:
            params: 超参数字典
            train_scores: 训练集分数历史
            val_scores: 验证集分数历史
            metric_name: 主要指标名称
            val_metrics: 验证集最终指标
            test_metrics: 测试集最终指标
            model: 训练好的模型（可选）
            model_name: 模型注册名称（可选，需要与 model 一起使用）
        """
        with mlflow.start_run():
            # 记录参数
            MLFlowUtils.log_params_batch(params)
            
            # 记录训练历史
            MLFlowUtils.log_training_history(train_scores, val_scores, metric_name)
            
            # 记录最终指标
            if val_metrics:
                MLFlowUtils.log_metrics_batch(val_metrics, prefix="val_")
            if test_metrics:
                MLFlowUtils.log_metrics_batch(test_metrics, prefix="test_")
            
            # 记录模型
            if model is not None and model_name is not None:
                MLFlowUtils.log_model_sklearn(model, model_name)
            
            logger.info("训练结果已完整记录到 MLflow")



from loguru import logger
from neco.ocr.base_ocr import BaseOCR


class PPOcr(BaseOCR):
    _ocr_engine = None

    @classmethod
    def get_engine(cls):
        if cls._ocr_engine is None:
            from paddleocr import PaddleOCR
            import paddle

            gpu_available = paddle.device.is_compiled_with_cuda()

            logger.info(f"初始化PaddleOCR,GPU可用性为:{gpu_available}")
            cls._ocr_engine = PaddleOCR(table=True, show_log=True,
                                        lang='ch', use_angle_cls=True,
                                        use_gpu=gpu_available,
                                        det_model_dir='models/pp_ocr_det',
                                        rec_model_dir='models/pp_ocr_rec',
                                        cls_model_dir='models/pp_ocr_cls')
        return cls._ocr_engine

    def __init__(self):
        self.ocr_engine = self.get_engine()

    def predict(self, file) -> str:
        logger.info(f"使用PaddleOCR识别图片:[{file}]")
        result = self.ocr_engine.ocr(file, cls=True)
        recognized_texts = ''
        if result:
            for lines in result:
                for line in lines:
                    recognized_texts += line[1][0] + ''
        return recognized_texts



from typing import Optional, Dict, Any
from langchain_openai import OpenAIEmbeddings
from loguru import logger
from singleton_decorator import singleton


@singleton
class EmbedManager:
    """嵌入模型管理器，支持本地和远程模型。"""
    
    # 协议前缀
    LOCAL_GPU_PREFIX = 'local-gpu:'
    LOCAL_PREFIX = 'local:'
    
    # 模型类型
    HUGGINGFACE_TYPE = 'huggingface_embedding'
    
    # 默认配置
    DEFAULT_TIMEOUT = 300

    def __init__(self):
        self._embed_instances: Dict[str, Any] = {}

    def _parse_protocol(self, protocol: str) -> tuple[str, str]:
        """解析协议字符串，返回模型类型和名称。"""
        parts = protocol.split(':')
        if len(parts) < 3:
            raise ValueError(
                f"Protocol format should be 'prefix:type:model_name', got: {protocol}")
        return parts[1], parts[2]

    def get_local_embed_instance(self, protocol: str, cache_folder: str = None) -> Optional[Any]:
        """获取本地嵌入模型实例。"""
        # 复用已存在的实例
        if protocol in self._embed_instances:
            return self._embed_instances[protocol]

        model_type, model_name = self._parse_protocol(protocol)
        
        # 确定设备类型
        device = 'cuda' if protocol.startswith(self.LOCAL_GPU_PREFIX) else 'cpu'
        logger.info(f"Creating local embed instance: {model_name} on {device}")

        if model_type != self.HUGGINGFACE_TYPE:
            logger.warning(f"Unsupported model type: {model_type}")
            return None

        try:
            from langchain_huggingface import HuggingFaceEmbeddings
            
            self._embed_instances[protocol] = HuggingFaceEmbeddings(
                model_name=model_name,
                model_kwargs={'device': device},
                encode_kwargs={'normalize_embeddings': False},
                cache_folder=cache_folder
            )
            return self._embed_instances[protocol]
        except Exception as e:
            logger.error(f"Failed to create local embed instance: {e}")
            raise

    def get_embed(self, protocol: str, model_name: str = '', 
                  model_api_key: str = '', model_base_url: str = '',
                  cache_folder: str = './models') -> Any:
        """获取嵌入模型实例。"""
        if not protocol:
            raise ValueError("Protocol cannot be empty")

        # 本地模型
        if protocol.startswith(self.LOCAL_PREFIX):
            return self.get_local_embed_instance(protocol, cache_folder)
        
        # 远程模型参数校验
        if not model_name or not model_api_key:
            raise ValueError("model_name and model_api_key are required for remote embedding")

        logger.info(f"Creating remote embed instance: {model_name}")
        return OpenAIEmbeddings(
            model=model_name,
            api_key=model_api_key,
            base_url=model_base_url,
            timeout=self.DEFAULT_TIMEOUT,
        )


from typing import List

from langchain_core.documents import Document

from neco.llm.chunk import BaseChunk


class FullChunk(BaseChunk):
    """全文分块器，不对文档进行分割，只添加元数据"""

    def _split_documents(self, docs: List[Document]) -> List[Document]:
        """
        不进行分割，直接返回原文档

        Args:
            docs: 待处理的文档列表

        Returns:
            原样返回的文档列表，不进行任何分割
        """
        return docs



from typing import List

from langchain_core.documents import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter
from loguru import logger

from neco.llm.chunk import BaseChunk


class RecursiveChunk(BaseChunk):
    """递归字符文本分割器，支持多种分隔符的智能分割"""

    def __init__(self, chunk_size: int = 500, chunk_overlap: int = 128):
        """
        初始化递归分块器

        Args:
            chunk_size: 每个块的目标大小
            chunk_overlap: 块之间的重叠大小
        """
        self.text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            length_function=len,
            is_separator_regex=False,
        )

    def _split_documents(self, docs: List[Document]) -> List[Document]:
        """使用递归字符分割器分割文档"""
        try:
            logger.info(f"开始递归分块处理，文档数量: {len(docs)}")
            result = self.text_splitter.split_documents(docs)
            logger.info(f"递归分块完成，分块数量: {len(result)}")
            return result
        except Exception as e:
            logger.error(f"递归分块处理失败: {str(e)}")
            raise



from typing import List, Optional
import time
import threading

import requests
from langchain_core.documents import Document
from loguru import logger

from neco.llm.rerank.rerank_config import ReRankConfig


class ReRankManager:
    _rerank_instance = {}
    _instance_lock = threading.Lock()

    # 配置常量
    DEFAULT_REMOTE_TIMEOUT = 10  # 远程调用超时时间（秒）
    MAX_RETRY_ATTEMPTS = 3  # 最大重试次数
    RETRY_DELAY_BASE = 1  # 重试延迟基数（秒）

    @classmethod
    def get_local_rerank_instance(cls, protocol: str):
        """
            local:bce:maidalun1020/bce-reranker-base_v1
        """
        if cls._rerank_instance.get(protocol, None) is None:
            with cls._instance_lock:
                # 双重检查锁定模式
                if cls._rerank_instance.get(protocol, None) is None:
                    model_type = protocol.split(':')[1]
                    model_name = protocol.split(':')[2]

                    if model_type == 'bce':
                        from neco.llm.rerank.bce_rerank import BCEReRank
                        cls._rerank_instance[protocol] = BCEReRank(
                            model_name_or_path=model_name)
                        logger.info(f"成功初始化本地重排序模型: {protocol}")
                    else:
                        logger.error(f"不支持的本地重排序模型类型: {model_type}")
                        return None

        return cls._rerank_instance[protocol]

    @staticmethod
    def rerank_documents(
            rerank_model_base_url: str,
            rerank_model_name: str,
            rerank_model_api_key: str,
            search_query: str,
            search_result: List[Document],
            rerank_top_k: int,
            threshold: Optional[float] = None
    ) -> List[Document]:
        """
        统一的文档重排序处理方法

        Args:
            rerank_model_base_url: 重排序模型的基础URL
            rerank_model_name: 重排序模型名称
            rerank_model_api_key: 重排序模型API密钥
            search_query: 搜索查询
            search_result: 待重排序的文档列表
            rerank_top_k: 返回的前K个文档
            threshold: 阈值过滤（可选，会除以10进行标准化）

        Returns:
            重排序并过滤后的文档列表
        """
        config = ReRankConfig(
            model_base_url=rerank_model_base_url,
            model_name=rerank_model_name,
            api_key=rerank_model_api_key,
            query=search_query,
            top_k=rerank_top_k,
            threshold=threshold
        )

        return ReRankManager.rerank_documents_with_config(config, search_result)

    @staticmethod
    def rerank_documents_with_config(config: ReRankConfig, search_result: List[Document]) -> List[Document]:
        """
        使用配置对象进行文档重排序

        Args:
            config: 重排序配置
            search_result: 待重排序的文档列表

        Returns:
            重排序并过滤后的文档列表
        """
        if not search_result:
            return search_result

        # 执行重排序
        reranked_docs = []
        if config.model_base_url.startswith('local:'):
            logger.info(f"使用本地ReRank模型进行重排序: {config.model_base_url}")
            reranked_docs = ReRankManager._handle_local_rerank(
                config, search_result)
        else:
            logger.info(f"使用远程ReRank模型进行重排序: {config.model_base_url}")
            reranked_docs = ReRankManager._handle_remote_rerank(
                config, search_result)

        # 应用阈值过滤（如果提供了阈值）
        if config.threshold is not None and config.threshold > 0:
            return ReRankManager._filter_by_threshold(reranked_docs, config.threshold)

        return reranked_docs

    @staticmethod
    def _filter_by_threshold(docs: List[Document], threshold: float) -> List[Document]:
        """
        根据阈值过滤文档

        Args:
            docs: 待过滤的文档列表
            threshold: 阈值（会除以10进行标准化）

        Returns:
            过滤后的文档列表
        """
        if not docs or threshold <= 0:
            return docs

        normalized_threshold = threshold / 10
        filtered_docs = []

        for doc in docs:
            # 优先使用 relevance_score，其次使用 _score
            score = doc.metadata.get(
                'relevance_score', doc.metadata.get('_score', 0))
            if score >= normalized_threshold:
                filtered_docs.append(doc)

        logger.info(
            f"阈值过滤: 原始文档数={len(docs)}, 过滤后文档数={len(filtered_docs)}, 阈值={normalized_threshold}")
        return filtered_docs

    @staticmethod
    def _handle_local_rerank(
            config: ReRankConfig,
            search_result: List[Document]
    ) -> List[Document]:
        """处理本地重排序模型"""
        try:
            # 提取文档内容
            passages = [doc.page_content for doc in search_result]

            # 获取重排序实例并执行重排序
            rerank_instance = ReRankManager.get_local_rerank_instance(
                config.model_base_url)
            if rerank_instance is None:
                logger.error(f"无法获取本地重排序实例: {config.model_base_url}")
                return search_result

            local_rerank_result = rerank_instance.rerank(
                config.query, passages)
            logger.info(
                f"本地重排序完成: 处理文档数={len(passages)}, 查询='{config.query[:50]}...' if len(config.query) > 50 else config.query")

            top_k_search_result = []
            top_rerank_ids = local_rerank_result['rerank_ids'][:config.top_k]
            rerank_scores = local_rerank_result['rerank_scores']

            for i in top_rerank_ids:
                if 0 <= i < len(search_result):
                    doc = search_result[i]
                    if hasattr(doc, 'metadata'):
                        # 设置相关性分数
                        if 0 <= i < len(rerank_scores):
                            doc.metadata['relevance_score'] = rerank_scores[i]
                        else:
                            doc.metadata['relevance_score'] = 0.0
                    top_k_search_result.append(doc)
                else:
                    logger.warning(f"无效的重排序索引 {i}，跳过")

            logger.info(
                f"本地重排序结果: 原始文档数={len(search_result)}, 重排序后文档数={len(top_k_search_result)}")
            return top_k_search_result

        except Exception as e:
            logger.error(f"本地重排序处理失败: {e}, 模型协议={config.model_base_url}")
            return search_result

    @staticmethod
    def _handle_remote_rerank(
            config: ReRankConfig,
            search_result: List[Document]
    ) -> List[Document]:
        """处理远程重排序模型"""
        headers = {
            "accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {config.api_key}"
        }

        rerank_content = [doc.page_content for doc in search_result]
        data = {
            "model": config.model_name,
            "query": config.query,
            "documents": rerank_content,
        }

        # 带重试的远程调用
        for attempt in range(ReRankManager.MAX_RETRY_ATTEMPTS):
            try:
                logger.info(
                    f"远程重排序调用尝试 {attempt + 1}/{ReRankManager.MAX_RETRY_ATTEMPTS}: {config.model_base_url}")
                response = requests.post(
                    config.model_base_url,
                    headers=headers,
                    json=data,
                    timeout=ReRankManager.DEFAULT_REMOTE_TIMEOUT
                )
                response.raise_for_status()
                rerank_api_result = response.json()

                if 'results' not in rerank_api_result:
                    logger.error(f"远程ReRank API响应格式错误: {rerank_api_result}")
                    return search_result

                rerank_result_items = rerank_api_result['results']

                # 筛选有效的重排序结果项
                valid_rerank_items = [
                    item for item in rerank_result_items
                    if 'relevance_score' in item and 'index' in item
                ]

                # 按相关性分数排序
                sorted_rerank_items = sorted(
                    valid_rerank_items, key=lambda x: x['relevance_score'], reverse=True
                )

                top_k_search_result = []
                for item in sorted_rerank_items[:config.top_k]:
                    original_index = item['index']
                    if 0 <= original_index < len(search_result):
                        doc = search_result[original_index]
                        if hasattr(doc, 'metadata'):
                            doc.metadata['relevance_score'] = item['relevance_score']
                        top_k_search_result.append(doc)
                    else:
                        logger.warning(f"无效的原始索引 {original_index}，跳过")

                logger.info(
                    f"远程重排序成功: 原始文档数={len(search_result)}, 重排序后文档数={len(top_k_search_result)}")
                return top_k_search_result

            except requests.exceptions.Timeout as e:
                logger.warning(
                    f"远程ReRank API超时 (尝试 {attempt + 1}/{ReRankManager.MAX_RETRY_ATTEMPTS}): {e}")
                if attempt < ReRankManager.MAX_RETRY_ATTEMPTS - 1:
                    delay = ReRankManager.RETRY_DELAY_BASE * \
                        (2 ** attempt)  # 指数退避
                    logger.info(f"等待 {delay} 秒后重试...")
                    time.sleep(delay)
                else:
                    logger.error(f"远程ReRank API调用最终失败，返回原始结果")
                    return search_result
            except requests.exceptions.RequestException as e:
                logger.error(
                    f"远程ReRank API请求失败 (尝试 {attempt + 1}/{ReRankManager.MAX_RETRY_ATTEMPTS}): {e}")
                if attempt < ReRankManager.MAX_RETRY_ATTEMPTS - 1:
                    delay = ReRankManager.RETRY_DELAY_BASE * \
                        (2 ** attempt)  # 指数退避
                    logger.info(f"等待 {delay} 秒后重试...")
                    time.sleep(delay)
                else:
                    logger.error(f"远程ReRank API调用最终失败，返回原始结果")
                    return search_result
            except (KeyError, TypeError) as e:
                logger.error(f"处理远程ReRank API响应时出错: {e}")
                return search_result
            except Exception as e:
                logger.error(f"远程重排序处理出现未知错误: {e}")
                return search_result



from typing import Optional
from loguru import logger

from neco.ocr.azure_ocr import AzureOCR
from neco.ocr.olm_ocr import OlmOcr
from neco.ocr.pp_ocr import PPOcr


class OcrManager:
    @classmethod
    def load_ocr(cls, ocr_type: str,
                 model: Optional[str] = None,
                 base_url: Optional[str] = None,
                 api_key: Optional[str] = None):
        ocr = None
        logger.debug(f"加载OCR服务，类型: {ocr_type}")

        if ocr_type == 'pp_ocr':
            logger.debug("初始化PP-OCR服务")
            ocr = PPOcr()

        if ocr_type == 'olm_ocr':
            logger.debug(f"初始化OLM-OCR服务，模型: {model}")
            ocr = OlmOcr(base_url=base_url, api_key=api_key, model=model)

        if ocr_type == 'azure_ocr':
            logger.debug(f"初始化Azure-OCR服务，endpoint: {base_url}")
            ocr = AzureOCR(api_key=api_key, endpoint=base_url)

        return ocr



from typing import List, Any

from langchain_core.documents import Document
from langchain_experimental.text_splitter import SemanticChunker
from loguru import logger

from neco.llm.chunk import BaseChunk


class SemanticChunk(BaseChunk):
    """语义分块器，基于语义相似度进行文档分割"""

    def __init__(self, semantic_embedding_model: Any, sentence_split_regex: str = r'(?<=[.?!。？！])\s*'):
        """
        初始化语义分块器

        Args:
            semantic_embedding_model: 用于计算语义相似度的嵌入模型
            sentence_split_regex: 句子分割正则表达式，默认按中英文标点符号分割
        """
        if semantic_embedding_model is None:
            raise ValueError("semantic_embedding_model 不能为空")

        self.semantic_embedding_model = semantic_embedding_model

        try:
            self.semantic_chunker = SemanticChunker(
                embeddings=semantic_embedding_model,
                sentence_split_regex=sentence_split_regex
            )
            logger.info(f"语义分块器初始化成功，使用正则表达式: {sentence_split_regex}")
        except Exception as e:
            logger.error(f"语义分块器初始化失败: {e}")
            raise

    def _split_documents(self, docs: List[Document]) -> List[Document]:
        """使用语义分块器分割文档"""
        if not docs:
            logger.warning("输入文档列表为空")
            return []

        try:
            logger.debug(f"开始语义分块，输入文档数量: {len(docs)}")
            result = self.semantic_chunker.split_documents(docs)
            logger.debug(f"语义分块完成，输出块数量: {len(result)}")
            return result
        except Exception as e:
            logger.error(f"语义分块失败: {e}")
            raise



from .base_chunk import BaseChunk
from .fixed_size_chunk import FixedSizeChunk
from .full_chunk import FullChunk
from .recursive_chunk import RecursiveChunk
from .semantic_chunk import SemanticChunk

__all__ = [
    "BaseChunk",
    "FixedSizeChunk", 
    "FullChunk",
    "RecursiveChunk",
    "SemanticChunk",
]


import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List

from langchain_core.documents import Document
from loguru import logger


class BaseChunk(ABC):
    """文档分块的基类，提供通用的元数据添加功能"""

    @abstractmethod
    def _split_documents(self, docs: List[Document]) -> List[Document]:
        """子类需要实现的文档分割方法"""
        pass

    def chunk(self, docs: List[Document]) -> List[Document]:
        """
        对文档进行分块，并添加必要的元数据

        Args:
            docs: 待分块的文档列表

        Returns:
            分块后的文档列表，包含chunk_id、chunk_number等元数据
        """
        if not docs:
            logger.warning("输入文档列表为空，直接返回空列表")
            return []

        logger.info(f"开始文档分块处理，文档数量: {len(docs)}")

        try:
            # 1. 调用子类的分割方法
            split_docs = self._split_documents(docs)
            logger.debug(f"文档分割完成，分块数量: {len(split_docs)}")

            # 2. 按照 segment_number 分组
            grouped_docs = defaultdict(list)
            for doc in split_docs:
                segment_number = doc.metadata.get('segment_number', 0)
                grouped_docs[segment_number].append(doc)

            logger.debug(f"文档分组完成，分组数量: {len(grouped_docs)}")

            # 3. 为每个组内的文档添加 chunk_number 和 chunk_id
            result_docs = []
            for segment_number, segment_docs in grouped_docs.items():
                for chunk_number, doc in enumerate(segment_docs):
                    doc.metadata['chunk_number'] = str(chunk_number)
                    doc.metadata['chunk_id'] = str(uuid.uuid4())
                    # 确保segment_number存在
                    doc.metadata['segment_number'] = segment_number
                    result_docs.append(doc)

            logger.info(f"文档分块处理完成，最终分块数量: {len(result_docs)}")
            return result_docs

        except Exception as e:
            logger.error(f"文档分块处理失败: {str(e)}", exc_info=True)
            raise



from typing import Optional

from pydantic import BaseModel, Field


class ReRankConfig(BaseModel):
    """重排序配置

    用于配置重排序模型的参数，支持本地和远程模型。
    本地模型格式：local:bce:model_name
    远程模型格式：http://example.com/rerank
    """

    model_base_url: str = Field(..., description="重排序模型的基础URL或本地协议")
    model_name: str = Field(..., description="重排序模型名称")
    api_key: str = Field(default="", description="API密钥，本地模型可为空")
    query: str = Field(..., description="查询文本")
    top_k: int = Field(..., gt=0, description="返回的前K个结果数量")
    threshold: Optional[float] = Field(
        default=None, ge=0.0, le=1.0, description="相关性分数阈值")



import base64
import requests
import json


class OlmOcr:
    def __init__(self, base_url: str, api_key: str, model="olmOCR-7B-0225-preview"):
        self.base_url = base_url
        self.api_key = api_key
        self.model = model

    def predict(self, file_path: str) -> str:
        # 读取图片并转换为base64编码
        with open(file_path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode('utf-8')

        # 构建请求体
        payload = {
            "model": self.model,
            "messages": [
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": """
                               你是一个具备OCR能力的智能体，阅读方式和人类一样。你可以从图片中提取文本信息。
                               下面是一张的图片，要求：
                                    1. 一步一步的思考
                                    2. 只需返回该文档的纯文本表示，就像您自然阅读一样
                                    3. 当识别出公式的时候，将公式转换为 LaTeX 格式，将表格转换为 Markdown 格式
                                    4. 不要产生幻觉
                            """
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{base64_image}"
                            }
                        }
                    ]
                }
            ],
            "temperature": 0.01,
            "max_tokens": 8000
        }

        # 发送请求
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}"
        }

        response = requests.post(
            f"{self.base_url}/chat/completions",
            headers=headers,
            data=json.dumps(payload)
        )

        # 解析响应
        if response.status_code == 200:
            result = response.json()
            # 提取文本内容
            if "choices" in result and len(result["choices"]) > 0:
                return result["choices"][0]["message"]["content"]
            else:
                return "无法识别文本"
        else:
            return f"请求失败，状态码: {response.status_code}, 响应: {response.text}"
